"""Python library to interact with eQ-3 Bluetooth Smart thermostats.

:copyright: (c) 2024-present Lennard Beers
:license: MIT, see LICENSE.md for more details.
"""

__title__ = "eq3btsmart"
__author__ = "Lennard Beers"
__license__ = "MIT"
__copyright__ = "Copyright (c) 2024-present"

from eq3btsmart.const import *
from eq3btsmart.exceptions import *
from eq3btsmart.models import *
from eq3btsmart.thermostat import *
